@@ -1,9 +1,7 @@
 from pkg_resources import get_distribution, DistributionNotFound
 
-try:
-    version = get_distribution(__name__).version
-except DistributionNotFound:
-    version = "unknown version"
+version = get_distribution(__name__).version
 
 from .offline import OfflineReader
 from .daq import DAQReader
+from .gseagen import GSGReader
