@@ -5,36 +5,35 @@ import km3io.definitions.trigger
 import km3io.definitions.fitparameters
 import km3io.definitions.reconstruction
 
+MAIN_TREE_NAME = "E"
 # 110 MB based on the size of the largest basket found so far in km3net
 BASKET_CACHE_SIZE = 110 * 1024**2
 
 
+class cached_property:
+    """A simple cache decorator for properties."""
+    def __init__(self, function):
+        self.function = function
+
+    def __get__(self, obj, cls):
+        if obj is None:
+            return self
+        prop = obj.__dict__[self.function.__name__] = self.function(obj)
+        return prop
+
+
 class OfflineKeys:
     """wrapper for offline keys"""
-    def __init__(self, file_path):
+    def __init__(self, tree):
         """OfflineKeys is a class that reads all the available keys in an offline
         file and adapts the keys format to Python format.
 
         Parameters
         ----------
-        file_path : path-like object
-            Path to the offline file of interest. It can be a str or any python
-            path-like object that points to the file of ineterst.
+        tree : uproot.TTree
+            The main ROOT tree.
         """
-        self._file_path = file_path
-        self._events_keys = None
-        self._hits_keys = None
-        self._tracks_keys = None
-        self._mc_hits_keys = None
-        self._mc_tracks_keys = None
-        self._valid_keys = None
-        self._fit_keys = None
-        self._cut_hits_keys = None
-        self._cut_tracks_keys = None
-        self._cut_events_keys = None
-        self._trigger = None
-        self._fitparameters = None
-        self._reconstruction = None
+        self._tree = tree
 
     def __str__(self):
         return '\n'.join([
@@ -46,10 +45,32 @@ class OfflineKeys:
         ])
 
     def __repr__(self):
-        return str(self)
-        # return f'{self.__class__.__name__}("{self._file_path}")'
+        return "<{}>".format(self.__class__.__name__)
 
-    @property
+    def _get_keys(self, tree, fake_branches=None):
+        """Get tree keys except those in fake_branches
+
+        Parameters
+        ----------
+        tree : uproot.Tree
+            The tree to look for keys
+        fake_branches : list of str or None
+            The fake branches to ignore
+
+        Returns
+        -------
+        list of str
+            The keys of the tree.
+        """
+        keys = []
+        for key in tree.keys():
+            key = key.decode('utf-8')
+            if fake_branches is not None and key in fake_branches:
+                continue
+            keys.append(key)
+        return keys
+
+    @cached_property
     def events_keys(self):
         """reads events keys from an offline file.
 
@@ -59,17 +80,12 @@ class OfflineKeys:
             list of all events keys found in an offline file,
             except those found in fake branches.
         """
-        if self._events_keys is None:
-            fake_branches = ['Evt', 'AAObject', 'TObject', 't']
-            t_baskets = ['t.fSec', 't.fNanoSec']
-            tree = uproot.open(self._file_path)['E']['Evt']
-            self._events_keys = [
-                key.decode('utf-8') for key in tree.keys()
-                if key.decode('utf-8') not in fake_branches
-            ] + t_baskets
-        return self._events_keys
+        fake_branches = ['Evt', 'AAObject', 'TObject', 't']
+        t_baskets = ['t.fSec', 't.fNanoSec']
+        tree = self._tree['Evt']
+        return self._get_keys(self._tree['Evt'], fake_branches) + t_baskets
 
-    @property
+    @cached_property
     def hits_keys(self):
         """reads hits keys from an offline file.
 
@@ -79,18 +95,10 @@ class OfflineKeys:
             list of all hits keys found in an offline file,
             except those found in fake branches.
         """
-        if self._hits_keys is None:
-            fake_branches = [
-                'hits.usr', 'hits.usr_names'
-            ]  # to be treated like trks.usr and trks.usr_names
-            tree = uproot.open(self._file_path)['E']['hits']
-            self._hits_keys = [
-                key.decode('utf8') for key in tree.keys()
-                if key.decode('utf8') not in fake_branches
-            ]
-        return self._hits_keys
+        fake_branches = ['hits.usr', 'hits.usr_names']
+        return self._get_keys(self._tree['hits'], fake_branches)
 
-    @property
+    @cached_property
     def tracks_keys(self):
         """reads tracks keys from an offline file.
 
@@ -100,20 +108,12 @@ class OfflineKeys:
             list of all tracks keys found in an offline file,
             except those found in fake branches.
         """
-        if self._tracks_keys is None:
-            # a solution can be tree['trks.usr_data'].array(
-            # uproot.asdtype(">i4"))
-            fake_branches = [
-                'trks.usr_data', 'trks.usr', 'trks.usr_names'
-            ]  # can be accessed using tree['trks.usr_names'].array()
-            tree = uproot.open(self._file_path)['E']['Evt']['trks']
-            self._tracks_keys = [
-                key.decode('utf8') for key in tree.keys()
-                if key.decode('utf8') not in fake_branches
-            ]
-        return self._tracks_keys
+        # a solution can be tree['trks.usr_data'].array(
+        # uproot.asdtype(">i4"))
+        fake_branches = ['trks.usr_data', 'trks.usr', 'trks.usr_names']
+        return self._get_keys(self._tree['Evt']['trks'], fake_branches)
 
-    @property
+    @cached_property
     def mc_hits_keys(self):
         """reads mc hits keys from an offline file.
 
@@ -123,16 +123,10 @@ class OfflineKeys:
             list of all mc hits keys found in an offline file,
             except those found in fake branches.
         """
-        if self._mc_hits_keys is None:
-            fake_branches = ['mc_hits.usr', 'mc_hits.usr_names']
-            tree = uproot.open(self._file_path)['E']['Evt']['mc_hits']
-            self._mc_hits_keys = [
-                key.decode('utf8') for key in tree.keys()
-                if key.decode('utf8') not in fake_branches
-            ]
-        return self._mc_hits_keys
+        fake_branches = ['mc_hits.usr', 'mc_hits.usr_names']
+        return self._get_keys(self._tree['Evt']['mc_hits'], fake_branches)
 
-    @property
+    @cached_property
     def mc_tracks_keys(self):
         """reads mc tracks keys from an offline file.
 
@@ -142,32 +136,23 @@ class OfflineKeys:
             list of all mc tracks keys found in an offline file,
             except those found in fake branches.
         """
-        if self._mc_tracks_keys is None:
-            fake_branches = [
-                'mc_trks.usr_data', 'mc_trks.usr', 'mc_trks.usr_names'
-            ]  # same solution as above can be used
-            tree = uproot.open(self._file_path)['E']['Evt']['mc_trks']
-            self._mc_tracks_keys = [
-                key.decode('utf8') for key in tree.keys()
-                if key.decode('utf8') not in fake_branches
-            ]
-        return self._mc_tracks_keys
+        fake_branches = [
+            'mc_trks.usr_data', 'mc_trks.usr', 'mc_trks.usr_names'
+        ]
+        return self._get_keys(self._tree['Evt']['mc_trks'], fake_branches)
 
-    @property
+    @cached_property
     def valid_keys(self):
         """constructs a list of all valid keys to be read from an offline event file.
         Returns
         -------
         list of str
             list of all valid keys.
-    """
-        if self._valid_keys is None:
-            self._valid_keys = (self.events_keys + self.hits_keys +
-                                self.tracks_keys + self.mc_tracks_keys +
-                                self.mc_hits_keys)
-        return self._valid_keys
+        """
+        return (self.events_keys + self.hits_keys + self.tracks_keys +
+                self.mc_tracks_keys + self.mc_hits_keys)
 
-    @property
+    @cached_property
     def fit_keys(self):
         """constructs a list of fit parameters, not yet outsourced in an offline file.
 
@@ -176,14 +161,11 @@ class OfflineKeys:
         list of str
             list of all "trks.fitinf" keys.
         """
-        if self._fit_keys is None:
-            self._fit_keys = sorted(self.fitparameters,
-                                    key=self.fitparameters.get,
-                                    reverse=False)
-            # self._fit_keys = [*fit.keys()]
-        return self._fit_keys
+        return sorted(km3io.definitions.fitparameters.data,
+                      key=km3io.definitions.fitparameters.data.get,
+                      reverse=False)
 
-    @property
+    @cached_property
     def cut_hits_keys(self):
         """adapts hits keys for instance variables format in a Python class.
 
@@ -192,13 +174,9 @@ class OfflineKeys:
         list of str
             list of adapted hits keys.
         """
-        if self._cut_hits_keys is None:
-            self._cut_hits_keys = [
-                k.split('hits.')[1].replace('.', '_') for k in self.hits_keys
-            ]
-        return self._cut_hits_keys
+        return [k.split('hits.')[1].replace('.', '_') for k in self.hits_keys]
 
-    @property
+    @cached_property
     def cut_tracks_keys(self):
         """adapts tracks keys for instance variables format in a Python class.
 
@@ -207,13 +185,11 @@ class OfflineKeys:
         list of str
             list of adapted tracks keys.
         """
-        if self._cut_tracks_keys is None:
-            self._cut_tracks_keys = [
-                k.split('trks.')[1].replace('.', '_') for k in self.tracks_keys
-            ]
-        return self._cut_tracks_keys
+        return [
+            k.split('trks.')[1].replace('.', '_') for k in self.tracks_keys
+        ]
 
-    @property
+    @cached_property
     def cut_events_keys(self):
         """adapts events keys for instance variables format in a Python class.
 
@@ -222,170 +198,64 @@ class OfflineKeys:
         list of str
             list of adapted events keys.
         """
-        if self._cut_events_keys is None:
-            self._cut_events_keys = [
-                k.replace('.', '_') for k in self.events_keys
-            ]
-        return self._cut_events_keys
-
-    @property
-    def trigger(self):
-        """trigger parameters and their index from km3net-Dataformat.
-
-        Returns
-        -------
-        dict
-            dictionary of trigger parameters and their index in an Offline
-            file.
-        """
-        if self._trigger is None:
-            self._trigger = km3io.definitions.trigger.data
-        return self._trigger
-
-    @property
-    def reconstruction(self):
-        """reconstruction parameters and their index from km3net-Dataformat.
-
-        Returns
-        -------
-        dict
-            dictionary of reconstruction parameters and their index in an
-            Offline file.
-        """
-        if self._reconstruction is None:
-            self._reconstruction = km3io.definitions.reconstruction.data
-        return self._reconstruction
-
-    @property
-    def fitparameters(self):
-        """fit parameters parameters and their index from km3net-Dataformat.
-
-        Returns
-        -------
-        dict
-            dictionary of fit parameters and their index in an Offline
-            file.
-        """
-        if self._fitparameters is None:
-            self._fitparameters = km3io.definitions.fitparameters.data
-        return self._fitparameters
-
-
-class Reader:
-    """Reader for one offline ROOT file"""
-    def __init__(self, file_path):
-        """ Reader class is an offline ROOT file reader. This class is a
-        "very" low level I/O.
-
-        Parameters
-        ----------
-        file_path : path-like object
-            Path to the file of interest. It can be a str or any python
-            path-like object that points to the file of ineterst.
-        """
-        self._file_path = file_path
-        self._data = uproot.open(self._file_path)['E'].lazyarrays(
-            basketcache=uproot.cache.ThreadSafeArrayCache(BASKET_CACHE_SIZE))
-        self._keys = None
-
-    def __getitem__(self, key):
-        """reads data stored in the branch of interest in an Evt tree.
-
-        Parameters
-        ----------
-        key : str
-            name of the branch of interest in event data.
-
-        Returns
-        -------
-        lazyarray
-            Lazyarray of all data stored in the branch of interest. A lazyarray
-            is an array-like object that reads data on demand. Here, only the
-            first and last chunks of data are read in memory, and not all data
-            in the array. The output can be used with all `Numpy's universal
-            functions <https://docs.scipy.org/doc/numpy/reference/ufuncs.html>`
-            .
-
-        Raises
-        ------
-        KeyError
-            Some branches in an offline file structure are "fake branches" and
-            do not contain data. Therefore, the keys corresponding to these
-            fake branches are not read.
-        """
-        keys = self.keys.valid_keys
-        if key not in keys and not isinstance(key, int):
-            raise KeyError(
-                "'{}' is not a valid key or is a fake branch.".format(key))
-        return self._data[key]
-
-    def __len__(self):
-        return len(self._data)
-
-    def __repr__(self):
-        return "<{}: {} entries>".format(self.__class__.__name__, len(self))
-
-    @property
-    def keys(self):
-        """wrapper for all keys in an offline file.
-
-        Returns
-        -------
-        Class
-            OfflineKeys.
-        """
-        if self._keys is None:
-            self._keys = OfflineKeys(self._file_path)
-        return self._keys
+        return [k.replace('.', '_') for k in self.events_keys]
 
 
 class OfflineReader:
     """reader for offline ROOT files"""
-    def __init__(self, file_path, data=None):
+    def __init__(self, file_path=None, fobj=None, data=None):
         """ OfflineReader class is an offline ROOT file wrapper
 
         Parameters
         ----------
         file_path : path-like object
             Path to the file of interest. It can be a str or any python
-            path-like object that points to the file of ineterst.
+            path-like object that points to the file.
+
         """
-        self._file_path = file_path
-        if data is not None:
-            self._data = data
-        else:
-            self._data = uproot.open(self._file_path)['E'].lazyarrays(
+        if file_path is not None:
+            self._fobj = uproot.open(file_path)
+            self._tree = self._fobj[MAIN_TREE_NAME]
+            self._data = self._tree.lazyarrays(
                 basketcache=uproot.cache.ThreadSafeArrayCache(
                     BASKET_CACHE_SIZE))
-        self._events = None
-        self._hits = None
-        self._tracks = None
-        self._mc_hits = None
-        self._mc_tracks = None
-        self._keys = None
-        self._best_reco = None
-        self._header = None
-        self._usr = None
+        else:
+            self._fobj = fobj
+            self._tree = self._fobj[MAIN_TREE_NAME]
+            self._data = data
 
-    def __getitem__(self, item):
-        return OfflineReader(file_path=self._file_path, data=self._data[item])
+    @classmethod
+    def from_index(cls, source, index):
+        """Create an instance with a subtree of a given index
+
+        Parameters
+        ----------
+        source: ROOTDirectory
+            The source file.
+        index: index or slice
+            The index or slice to create the subtree.
+        """
+        instance = cls(fobj=source._fobj, data=source._data[index])
+        return instance
+
+    def __getitem__(self, index):
+        return OfflineReader.from_index(source=self, index=index)
 
     def __len__(self):
         return len(self._data)
 
-    @property
+    @cached_property
     def header(self):
-        if self._header is None:
-            fobj = uproot.open(self._file_path)
-            if 'Head' in fobj:
-                self._header = {}
-                for n, x in fobj['Head']._map_3c_string_2c_string_3e_.items():
-                    self._header[n.decode("utf-8")] = x.decode("utf-8").strip()
-            else:
-                warnings.warn("Your file header has an unsupported format")
-        return self._header
+        if 'Head' in self._fobj:
+            header = {}
+            for n, x in self._fobj['Head']._map_3c_string_2c_string_3e_.items(
+            ):
+                header[n.decode("utf-8")] = x.decode("utf-8").strip()
+            return header
+        else:
+            warnings.warn("Your file header has an unsupported format")
 
-    @property
+    @cached_property
     def keys(self):
         """wrapper for all keys in an offline file.
 
@@ -394,11 +264,9 @@ class OfflineReader:
         Class
             OfflineKeys.
         """
-        if self._keys is None:
-            self._keys = OfflineKeys(self._file_path)
-        return self._keys
+        return OfflineKeys(self._tree)
 
-    @property
+    @cached_property
     def events(self):
         """wrapper for offline events.
 
@@ -407,13 +275,11 @@ class OfflineReader:
         Class
             OfflineEvents.
         """
-        if self._events is None:
-            self._events = OfflineEvents(
-                self.keys.cut_events_keys,
-                [self._data[key] for key in self.keys.events_keys])
-        return self._events
+        return OfflineEvents(
+            self.keys.cut_events_keys,
+            [self._data[key] for key in self.keys.events_keys])
 
-    @property
+    @cached_property
     def hits(self):
         """wrapper for offline hits.
 
@@ -422,13 +288,10 @@ class OfflineReader:
         Class
             OfflineHits.
         """
-        if self._hits is None:
-            self._hits = OfflineHits(
-                self.keys.cut_hits_keys,
-                [self._data[key] for key in self.keys.hits_keys])
-        return self._hits
+        return OfflineHits(self.keys.cut_hits_keys,
+                           [self._data[key] for key in self.keys.hits_keys])
 
-    @property
+    @cached_property
     def tracks(self):
         """wrapper for offline tracks.
 
@@ -437,14 +300,11 @@ class OfflineReader:
         Class
             OfflineTracks.
         """
-        if self._tracks is None:
-            self._tracks = OfflineTracks(
-                self.keys.cut_tracks_keys,
-                [self._data[key] for key in self.keys.tracks_keys],
-                fitparameters=self.keys.fitparameters)
-        return self._tracks
+        return OfflineTracks(
+            self.keys.cut_tracks_keys,
+            [self._data[key] for key in self.keys.tracks_keys])
 
-    @property
+    @cached_property
     def mc_hits(self):
         """wrapper for offline mc hits.
 
@@ -453,13 +313,10 @@ class OfflineReader:
         Class
             OfflineHits.
         """
-        if self._mc_hits is None:
-            self._mc_hits = OfflineHits(
-                self.keys.cut_hits_keys,
-                [self._data[key] for key in self.keys.mc_hits_keys])
-        return self._mc_hits
+        return OfflineHits(self.keys.cut_hits_keys,
+                           [self._data[key] for key in self.keys.mc_hits_keys])
 
-    @property
+    @cached_property
     def mc_tracks(self):
         """wrapper for offline mc tracks.
 
@@ -468,21 +325,15 @@ class OfflineReader:
         Class
             OfflineTracks.
         """
-        if self._mc_tracks is None:
-            self._mc_tracks = OfflineTracks(
-                self.keys.cut_tracks_keys,
-                [self._data[key] for key in self.keys.mc_tracks_keys],
-                fitparameters=self.keys.fitparameters)
-        return self._mc_tracks
+        return OfflineTracks(
+            self.keys.cut_tracks_keys,
+            [self._data[key] for key in self.keys.mc_tracks_keys])
 
-    @property
+    @cached_property
     def usr(self):
-        if self._usr is None:
-            self._usr = Usr(self._file_path)
-        return self._usr
+        return Usr(self._tree)
 
-    @property
-    def best_reco(self):
+    def get_best_reco(self):
         """returns the best reconstructed track fit data. The best fit is defined
         as the track fit with the maximum reconstruction stages. When "nan" is
         returned, it means that the reconstruction parameter of interest is not
@@ -496,24 +347,22 @@ class OfflineReader:
         numpy recarray
             a recarray of the best track fit data (reconstruction data).
         """
-        if self._best_reco is None:
-            keys = ", ".join(self.keys.fit_keys[:-1])
-            empty_fit_info = np.array(
-                [match for match in self._find_empty(self.tracks.fitinf)])
-            fit_info = [
-                i for i, j in zip(self.tracks.fitinf, empty_fit_info[:, 1])
-                if j is not None
-            ]
-            stages = self._get_max_reco_stages(self.tracks.rec_stages)
-            fit_data = np.array([i[j] for i, j in zip(fit_info, stages[:, 2])])
-            rows_size = len(max(fit_data, key=len))
-            equal_size_data = np.vstack([
-                np.hstack([i, np.zeros(rows_size - len(i)) + np.nan])
-                for i in fit_data
-            ])
-            self._best_reco = np.core.records.fromarrays(
-                equal_size_data.transpose(), names=keys)
-        return self._best_reco
+        keys = ", ".join(self.keys.fit_keys[:-1])
+        empty_fit_info = np.array(
+            [match for match in self._find_empty(self.tracks.fitinf)])
+        fit_info = [
+            i for i, j in zip(self.tracks.fitinf, empty_fit_info[:, 1])
+            if j is not None
+        ]
+        stages = self._get_max_reco_stages(self.tracks.rec_stages)
+        fit_data = np.array([i[j] for i, j in zip(fit_info, stages[:, 2])])
+        rows_size = len(max(fit_data, key=len))
+        equal_size_data = np.vstack([
+            np.hstack([i, np.zeros(rows_size - len(i)) + np.nan])
+            for i in fit_data
+        ])
+        return np.core.records.fromarrays(equal_size_data.transpose(),
+                                          names=keys)
 
     def _get_max_reco_stages(self, reco_stages):
         """find the longest reconstructed track based on the maximum size of 
@@ -825,15 +674,13 @@ class OfflineReader:
 
 class Usr:
     """Helper class to access AAObject usr stuff"""
-    def __init__(self, filepath):
-        self._f = uproot.open(filepath)
+    def __init__(self, tree):
         # Here, we assume that every event has the same names in the same order
         # to massively increase the performance. This needs triple check if it's
         # always the case; the usr-format is simply a very bad design.
         try:
             self._usr_names = [
-                n.decode("utf-8")
-                for n in self._f['E']['Evt']['usr_names'].array()[0]
+                n.decode("utf-8") for n in tree['Evt']['usr_names'].array()[0]
             ]
         except (KeyError, IndexError):  # e.g. old aanet files
             self._usr_names = []
@@ -842,7 +689,7 @@ class Usr:
                 name: index
                 for index, name in enumerate(self._usr_names)
             }
-            self._usr_data = self._f['E']['Evt']['usr'].lazyarray(
+            self._usr_data = tree['Evt']['usr'].lazyarray(
                 basketcache=uproot.cache.ThreadSafeArrayCache(
                     BASKET_CACHE_SIZE))
             for name in self._usr_names:
@@ -918,9 +765,6 @@ class OfflineEvent:
             for k, v in zip(self._keys, self._values)
         ])
 
-    def __repr__(self):
-        return str(self)
-
 
 class OfflineHits:
     """wrapper for offline hits"""
@@ -982,19 +826,10 @@ class OfflineHit:
     def __getitem__(self, item):
         return self._values[item]
 
-    def __repr__(self):
-        return str(self)
-
-    # def _is_empty(array):
-    #     if array.size:
-    #         return False
-    #     else:
-    #         return True
-
 
 class OfflineTracks:
     """wrapper for offline tracks"""
-    def __init__(self, keys, values, fitparameters=None):
+    def __init__(self, keys, values):
         """wrapper for offline tracks
 
         Parameters
@@ -1003,20 +838,14 @@ class OfflineTracks:
             list of cropped tracks keys.
         values : list of arrays
             list of arrays containting tracks data.
-        fitparameters : None, optional
-            dictionary of tracks fit information (not yet outsourced in offline
-            files).
         """
         self._keys = keys
         self._values = values
-        if fitparameters is not None:
-            self._fitparameters = fitparameters
         for k, v in zip(self._keys, self._values):
             setattr(self, k, v)
 
     def __getitem__(self, item):
-        return OfflineTrack(self._keys, [v[item] for v in self._values],
-                            fitparameters=self._fitparameters)
+        return OfflineTrack(self._keys, [v[item] for v in self._values])
 
     def __len__(self):
         try:
@@ -1034,7 +863,7 @@ class OfflineTracks:
 
 class OfflineTrack:
     """wrapper for an offline track"""
-    def __init__(self, keys, values, fitparameters=None):
+    def __init__(self, keys, values):
         """wrapper for one offline track.
 
         Parameters
@@ -1043,14 +872,9 @@ class OfflineTrack:
             list of cropped tracks keys.
         values : list of arrays
             list of arrays containting track data.
-        fitparameters : None, optional
-            dictionary of tracks fit information (not yet outsourced in offline
-            files).
         """
         self._keys = keys
         self._values = values
-        if fitparameters is not None:
-            self._fitparameters = fitparameters
         for k, v in zip(self._keys, self._values):
             setattr(self, k, v)
 
@@ -1061,12 +885,9 @@ class OfflineTrack:
         ]) + "\n\t" + "\n\t".join([
             "{:30} {:^2} {:>26}".format(k, ':', str(
                 getattr(self, 'fitinf')[v]))
-            for k, v in self._fitparameters.items()
+            for k, v in km3io.definitions.fitparameters.data.items()
             if len(getattr(self, 'fitinf')) > v
         ])  # I don't like 18 being explicit here
 
     def __getitem__(self, item):
         return self._values[item]
-
-    def __repr__(self):
-        return str(self)
